unit Unit1;

{  freehand drawing
   record up to 9 drawings

   drawing recorded as 4 bit quantities (strokes) and stored in
   word array on heap.

   format:  buffer[0] holds number of strokes
            buffer[1..] hold strokes 0 1 2 3 , 4 5 6 7 , .........
            stroke coding:

     stroke code   0  1  2  3  4  5  6  7  8  9  10 11 12 13 14 15
     next X        1  1  0 -1 -1 -1  0  1  2  2   0 -2 -2 -2  0  2
     next Y        0 -1 -1 -1  0  1  1  1  0 -2  -2 -2  0  2  2  2
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Buttons, ExtCtrls, dav7components;

type
  TForm1 = class(TForm)
    PaintBox1: TPaintBox;
    dav7ColorPicker1: Tdav7ColorPicker;
    Label1: TLabel;
    DavArrayBtn1: TDavArrayBtn;
    Label2: TLabel;
    BitBtn1: TBitBtn;
    BitBtn2: TBitBtn;
    BitBtn3: TBitBtn;
    procedure FormPaint(Sender: TObject);
    procedure DavArrayBtn1BtnPaint(sender: TObject; BtnNr: Byte;
      status: TBtnStatus);
    procedure FormCreate(Sender: TObject);
    procedure dav7ColorPicker1Select(sender: TObject; color: Integer);
    procedure PaintBox1Paint(Sender: TObject);
    procedure DavArrayBtn1BtnChange(sender: TObject; BtnNr: Byte;
      status: TBtnStatus; button: TMouseButton);
    procedure PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintBox1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormDestroy(Sender: TObject);
    procedure BitBtn3Click(Sender: TObject);
    procedure BitBtn1Click(Sender: TObject);
    procedure BitBtn2Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

const strokesInBuf = 5000;
      strokeWords = strokesInBuf div 4;

type TA = array[0..strokewords] of WORD;
     PA = ^TA;

     freeprops = record                 //properties of free drawing
                  prColor : DWORD;
                  prPenwidth : byte;
                  startX,startY : DWORD;
                 end;

var buffer : array[1..9] of PA; //mem pointers 1..9
    bufNr   : byte = 0;         //1..9
    props : array[1..9] of freeprops;

    pencolor : DWORD = $00000000; //color selection
    penwidth : byte  = 2;         //pen width selection

    drawing  : boolean = false;   //true if drawing in progress

    x1,y1,x2,y2 : integer;  //line begin..end
    oldstroke : byte;
    newstrdir : boolean;

procedure debug(nr : byte);
//display strokes in 4 columns of 25
var p : PA;
    w,i,n : word;
    s,str : byte;
    row,column : byte;
begin
 with form1.paintbox1 do with canvas do
  begin
   font.name := 'arial';
   font.height := 18;
   p := buffer[nr];
   n := p^[0];
   if n > 100 then n := 100;
   for i := 0 to n-1 do
    begin
     w := 1 + (i shr 2);
     s := (i and $3) shl 2;
     str := (p^[w] shr s) and $f;
     column := i div 25;
     row := i mod 25;
     textout(column*50 + 600,row*20,inttostr(str));
    end;
  end;
end;

//--------- painting an already recorded drawing ------------
  
procedure freepaint(nr :byte);
//paint strokes in buffer
var snr : word;  //stroke nr
    str : shortInt;
    w   : word;  //in array
    s   : byte;  //stroke shift count in word
    x,y : integer;
    p : PA;
begin
 p := buffer[Nr];
 with form1.paintbox1.canvas do
  begin
   with props[Nr] do
    begin
     x := startX;
     y := startY;
     pen.color := prColor;
     pen.Width := prPenWidth;
    end;
   moveto(x,y);
   for snr := 0 to p^[0]-1 do
    begin
     w := 1 + (snr shr 2);
     s := (snr and $3) shl 2;//shift count 0,4,8,12
     str := (p^[w] shr s) and $f;
     while str >= 0 do
      begin
       case str of
        0,8  : inc(x);
        1,9  : begin
                inc(x); dec(y);
               end;
        2,10 : dec(y);
        3,11 : begin
                dec(x); dec(y);
               end;
        4,12 : dec(x);
        5,13 : begin
                dec(x); inc(y);
               end;
        6,14 : inc(y);
        7,15 : begin
                inc(x); inc(y);
               end;
       end;//case
       dec(str,8);
       lineto(x,y);
      end;//while
    end;//with
  end;//for
end;

// ------- recording a drawing -----------------

function requestbufferspace : boolean;
//request heap space to record the drawing
//return "true" if OK
//store pointer in buffer
begin
 if bufNr = 9 then
  begin
   result := false;
   exit;
  end;

 inc(bufNr);
 try
  Getmem(buffer[bufNr],2*(strokewords+1));
  result := true;
 except
  result := false;
 end;
end;

function sign(v : single) : integer;
//return 1 for positive, -1 for negative number
begin
 if v < 0 then result := -1 else result := 1;
end;

//=====

procedure savestroke(str : byte; var saved : boolean);
//store stroke str in buffer[bufNr]
//if possible, combine with previous stroke
//report in "saved" : true=OK
var p : PA;
    n,n1,mask : word;  //stroke number  0..4999
    w : word;          //word of stroke 1..4000 in array
    s : byte;          //shift count for stroke position
begin
 p := buffer[bufNr];
 n := p^[0];
 if n = strokesInBuf - 1 then
  begin
   saved := false;
   exit;
  end;

 saved := true;

 if (n = 0) then
  begin
   oldstroke := 0;
   newstrdir := true;
  end
   else
    begin
     n1 := n - 1;                   //n1: old stroke#
     w := 1 + (n1 shr 2);           //1+ : [0] holds stroke count
     s := (n1 and $3) shl 2;        //s = 0,4,8,12
     if (oldstroke = str) then
      begin
       p^[w] := p^[w] or (8 shl s);  //add bit 3 for double move
       newstrdir := false;
       oldstroke := oldstroke or $8;
       exit;
      end;
     if newstrdir then
      begin
       case (oldstroke shl 4) or str of
        $02,$20 : str := $f1;
        $06,$60 : str := $f7;
        $46,$64 : str := $f5;
        $24,$42 : str := $f3;
       end;//case
       if str > $f0 then
        begin
         str := str and $7;
         mask := $ffff xor ($f shl s);
         p^[w] := (p^[w] and mask) or (str shl s);
         oldstroke := str;
         exit;
        end;
      end;//if newstr
    end;

  w := 1 + (n shr 2);
  s := (n and $3) shl 2;
  mask := $ffff xor ($f shl s);
  p^[w] := (p^[w] and mask) or (str shl s);
  inc(n);
  p^[0] := n;
  newstrdir := oldstroke <> str; //== new code end
  oldstroke := str;
end;

function recordstrokes : boolean;
//called by mouse move
//translate mouse movement (x2,y2)--> (x3,y3) into strokes and store in buffer
const dir2stroke: array[0..15] of byte =  (0,0,0,4,0,0,0,0,6,0,7,5,2,0,1,3);
var rx,ry,dx,dy : single;
    dir,stroke,i : byte;   //direction
    steps,px1,py1,px2,py2 : integer;
begin
 dx := x2 - x1;
 dy := y2 - y1;
 if abs(dx) > abs(dy) then
  begin
   steps := trunc(abs(dx));
   dy := dy/steps; dx := sign(dx);      //horizontal orientation
  end
  else begin
        steps := trunc(abs(dy));
        dx := dx/steps; dy := sign(dy); //vertical orientation
       end;
 px1 := x2; py1 := y2;
 rx := x2; ry := y2;
 for i := 1 to steps do
  begin
   rx := rx + dx; ry := ry + dy;
   px2 := trunc(rx+0.5);
   py2 := trunc(ry+0.5);
   dir := 0;
   if px2 > px1 then dir := dir or 2;
   if px2 < px1 then dir := dir or 3;
   if py2 > py1 then dir := dir or 8;
   if py2 < py1 then dir := dir or $c;
   stroke := dir2stroke[dir];
   savestroke(stroke,result);
   if result = false then exit;  //buffer full

   px1 := px2; py1 := py2;
  end;
end;

procedure clearpaintbox;
begin
 with form1.paintbox1 do with canvas do
  begin
   brush.Color := $e0ffff;
   brush.Style := bsSolid;
   fillrect(rect(0,0,width,height));
  end;
end;

procedure repaintAll;
//paint background of paintbox1 and all free drawings
var i : byte;
begin
 clearpaintbox;
 for i := 1 to 9 do
  if assigned(buffer[i]) then freepaint(i);
end;

// ------- property selection --------------

procedure TForm1.DavArrayBtn1BtnChange(sender: TObject; BtnNr: Byte;
  status: TBtnStatus; button: TMouseButton);
  //set new pen width
begin
 if status = stDown then penwidth := btnNr + 1;
end;

procedure TForm1.dav7ColorPicker1Select(sender: TObject; color: Integer);
//a new color was selected
begin
 pencolor := color;
 davarraybtn1.Invalidate;
end;

//-------- mouse events ----------------

procedure TForm1.PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
  var p : PA;
begin
 if requestbufferspace then
  begin
   p := buffer[bufNr];
   p^[0] := 0;          //no strokes recorded yet
   p^[1] := 0;          //oldstrokes = 0
   drawing := true;
   x1 := x; y1 := y;
   with props[bufNr] do           //save properties for repaint
    begin
     prColor := penColor;
     prPenWidth := penwidth;
     startX := x;
     startY := y;
    end;
   with paintbox1.Canvas do      //set canvas
    begin
     pen.Width := penwidth;
     pen.Color := penColor;
    end;   
  end
  else showmessage('no buffer space');
end;

procedure TForm1.PaintBox1MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
  var p : PA;
//stop drawing
//if no strokes recorded, free buffer space
begin
 drawing := false;
 if (bufNr <> 0) then
  begin
   p := buffer[bufNr];
   if p^[0] = 0 then      //if no strokes recorded
    begin
     freemem(p);
     buffer[bufNr] := nil;
     dec(Bufnr);
    end
    else reAllocMem(p,((p^[0]+1) shr 1)+2); //trim buffer to needed space
//   debug(bufnr); //====
  end;
end;

procedure TForm1.PaintBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
 with paintbox1 do with canvas do
  begin
   if (x < 0) or (y < 0) or (x >= width) or (y >= height) then exit;

   if drawing then
    if (x <> x1) or (y <> y1) then
     begin
      x2 := x; y2 := y;
      moveto(x1,y1);
      lineto(x2,y2);
      if recordstrokes then begin
                             x1 := x; y1:= y; //save old position
                            end
      else begin
            showmessage('buffer full');
            drawing := false;
           end;
     end;
  end;//with..
end;

//--- create and destroy -----------

procedure TForm1.FormCreate(Sender: TObject);
//set all buffer pointers to nil
var i : byte;
begin
 for i := 1 to 9 do buffer[i] := nil;
 davarraybtn1.BtnDown(1);  //select default width = 2
end;

procedure TForm1.FormDestroy(Sender: TObject);
//free memory when closing application
var i : byte;
begin
 for i := 1 to 9 do
  if assigned(buffer[i]) then freemem(buffer[i]);
end;

//--- buttons ----

procedure TForm1.BitBtn3Click(Sender: TObject);
//repaint button pressed
begin
 repaintall;
end;

procedure TForm1.BitBtn1Click(Sender: TObject);
//clear button pressed
begin
 clearpaintbox;
end;

procedure TForm1.BitBtn2Click(Sender: TObject);
//backspace, remove last free drawing
//repaint
begin
 if bufNr > 0 then
  begin
   freemem(buffer[bufNr]);
   buffer[bufNr] := nil;
   dec(bufNr);
   repaintall;
  end;
end;

//----- paints --------

procedure TForm1.FormPaint(Sender: TObject);
//paint edge around paintbox1
var x1,y1,x2,y2 : integer;
    i : byte;
begin
 with paintbox1 do
  begin
   x1 := left-2; x2 := left+width+1;
   y1 := top-2;  y2 := top + height+1;
  end;
 with canvas do       //is form1.canvas
  begin
   pen.Width := 1;
   for i := 0 to 1 do
    begin
     pen.Color := $80ff80;
     moveto(x2-i,y1+i);
     lineto(x1+i,y1+i);
     lineto(x1+i,y2-i);
     pen.color := $008000;
     lineto(x2-i,y2-i);
     lineto(x2-i,y1+i);
    end;
  end;
end;

procedure TForm1.DavArrayBtn1BtnPaint(sender: TObject; BtnNr: Byte;
  status: TBtnStatus);
//paint the penwidth image for menu selection
var r : Trect;
    y : integer;
begin
 with davArrayBtn1 do with canvas do
  begin
   pen.Width := BtnNr + 1;
   pen.Color := pencolor;
   r := getBtnRect(BtnNr);
   y := r.Top + btnheight div 2;
   moveto(r.Left+5,y);
   lineto(r.Right-5,y);
  end;
end;

procedure TForm1.PaintBox1Paint(Sender: TObject);
//paint background & all buffers in use
begin
 repaintAll;
end;

end.
